/*\file CGAL_MSDM.cpp
\brief Implementation of the gateway between MATLAB and C++

\author Guillaume Lavou&eacute;, Xavier Rolland-Nevi&egrave;re
*/

#include "CGAL_MSDM.h"

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
	if (nrhs < 4)
	{
		mexErrMsgIdAndTxt("CGAL_MSDM:argChk",
			"Expected syntax is CGAL_MSDM(V1, F1, V2, F2, radius) where:\n" 
			"* V1, V2 are 3 by N matrices of vertex positions\n"
			"* F1, F2 are 3 by M matrices with the vertex indices (1 to N) in every facet\n"
			"* [optional] options structure with fields:\n"
			"** display (true) Display information in MATLAB console\n"
			"** multithreading (true)\n"
			"** nb_threads (max - 1)\n"
			"** radius (0.005) size of the local neighborhood\n"
			"** isGeodesic (true) geodesic distances in the neighborhood definition\n"
			"Optional output: Local MSDM values\n");
	}	
	if (mxGetM(prhs[0]) != 3)
	{
		mexErrMsgIdAndTxt("CGAL_MSDM:argChk","V1 must be 3 by N");
	}	
	if (mxGetM(prhs[1]) != 3)
	{
		mexErrMsgIdAndTxt("CGAL_MSDM:argChk","F1 must be 3 by N");
	}
	if (mxGetM(prhs[2]) != 3)
	{
		mexErrMsgIdAndTxt("CGAL_MSDM:argChk","V2 must be 3 by N");
	}
	if (mxGetM(prhs[3]) != 3)
	{
		mexErrMsgIdAndTxt("CGAL_MSDM:argChk","F2 must be 3 by N");
	}
	auto pP1 = Utils::MATLAB2Polyhedron(prhs[0], prhs[1]);
	auto pP2 = Utils::MATLAB2Polyhedron(prhs[2], prhs[3]);
	MSDM_options options;
	init_MSDM_options(options);
	if (nrhs == 5)
		parse_MSDM_options(prhs[4], options);
	std::vector<double> local_MSDM;
	double MSDM = compute_MSDM(*pP1, *pP2, options, local_MSDM);
	plhs[0] = mxCreateDoubleScalar(MSDM);
	if (nlhs == 2)
	{
		plhs[1] = mxCreateDoubleMatrix(1, pP1->size_of_vertices(), mxREAL);
		std::copy(local_MSDM.begin(), local_MSDM.end(), mxGetPr(plhs[1]));
	}
}

double compute_MSDM(Polyhedron& p1, Polyhedron& p2, const MSDM_options& options, std::vector<double>& local_MSDM)
{	
	p1.construct();
	p2.construct();
	mwSize failures = NormalCycle::principal_curvatures(p1, options);
	mwSize failures2 = NormalCycle::principal_curvatures(p2, options);
	Distance3D m_pDistance(&p1, &p2, options);
	mwSize failures3 = m_pDistance.buildPairwiseEstimation();
	if (options.display && failures != 0)
		mexPrintf("%d failures\n", failures3);
	return m_pDistance.updateMSDM(local_MSDM);
}

void init_MSDM_options(MSDM_options& options)
{
	options.multithreading = true;
#pragma omp parallel
	{
#pragma omp master
		{
			options.nb_threads = (omp_get_num_threads() > 1) ? (omp_get_num_threads() - 1) : 1;
		}
	}
	options.radius = 0.015;
	options.isGeodesic = true;
	options.display = true;
	options.alpha = 0.4;
	options.beta = 0.4;
	options.gamma = 0.2;
	options.Mpower = 3;
	options.isRelative = true;
}

void parse_MSDM_options(const mxArray* MATLAB_MSDM_options, MSDM_options& options)
{
	getMatlabFieldScalar(MATLAB_MSDM_options, 0, "multithreading", &options.multithreading);
	getMatlabFieldScalar(MATLAB_MSDM_options, 0, "nb_threads", &options.nb_threads);
	getMatlabFieldScalar(MATLAB_MSDM_options, 0, "display", &options.display);
	getMatlabFieldScalar(MATLAB_MSDM_options, 0, "isGeodesic", &options.isGeodesic);
	getMatlabFieldScalar(MATLAB_MSDM_options, 0, "radius", &options.radius);
	getMatlabFieldScalar(MATLAB_MSDM_options, 0, "alpha", &options.alpha);
	getMatlabFieldScalar(MATLAB_MSDM_options, 0, "beta", &options.beta);
	getMatlabFieldScalar(MATLAB_MSDM_options, 0, "gamma", &options.gamma);
	getMatlabFieldScalar(MATLAB_MSDM_options, 0, "Mpower", &options.Mpower);
	getMatlabFieldScalar(MATLAB_MSDM_options, 0, "isRelative", &options.isRelative);
}