#ifndef DISTANCE_3D_H
#define DISTANCE_3D_H

#include "Header.h"
#include <set>
#include <stack>

/*! \brief Estimation of the MSDM between 2 polyhedron with curvature information
\copyright LIRIS, M2DisCo 2009, Licence QPL
\author Guillaume Lavou&eacute;, Xavier Rolland-Nevi&egrave;re

According to: <i>Perceptually driven 3D distance metrics with application to watermarking</i>, 
Lavou&eacute;, G., Drelie Gelasca, E., Dupont, F., Baskurt, A., Ebrahimi, T., 
in Proceedings of SPIE Applications of Digital Image Processing XXIX, August 2006.
*/
class Distance3D
{
private:
	Polyhedron *m_PolyOriginal, *m_PolyDegrad;
	double mp_wi_alpha, mp_wi_beta, mp_radius;
	MSDM_options mOptions;
private:

	void vertex_curvature_average_variance(Polyhedron* PolyUsed, Polyhedron::Vertex_handle pVertex, 
		std::vector<double> &TabDistance, std::vector<Point> &TabPoint);

	/*! \brief Estimate S between every pair of vertices (attacked/original). */
	void vertex_curvature_covariance(Polyhedron::Vertex_handle pVertex, Polyhedron::Vertex_handle pVertexDeg, 
		const std::vector<double>& TabDistance, const std::vector<Point>& TabPoint, 
		const std::vector<double>& TabDistanceDeg, const std::vector<Point>& TabPointDeg);

	inline double weighted_covariance(const std::vector<double>& TabD, const std::vector<double>& TabD2, 
		int ecart, const std::vector<Point>& TabP, const std::vector<Point>& TabP2, 
		const Point& O, double moyenne, double moyenneDeg);

public:	

	/*! \brief Constructor of the MSDM computation 
		\param _PolyOriginal Pointer to the original polyhedron
		\param _PolyDegrad Pointer to the altered polyhedron
		\param options Parameter settings of the MSDM
	*/
	Distance3D(Polyhedron *_PolyOriginal, Polyhedron *_PolyDegrad, const MSDM_options& options);

	/*! \brief Compute local MSDM at all vertices and MSDM from previous computations.
		\param [out] lMSDM container of local MSDM values
		\return MSDM

		Based on the settings in MSDM_options, compute the weighted sum of curvature variations at every vertex (local MSDM)
		and return the global Minkowski sum (MSDM). This procedure is multithreaded.
	*/
	double updateMSDM(std::vector<double>& lMSDM) const;

	/*! \brief Build estimation of L, C and S at every vertex. 
		\return Number of failures
	
		<b>This procedure should be called prior to updateMSDM and assumes that both versions of the polyhedron have
		curvature estimates.</b>
	*/
	mwSize buildPairwiseEstimation();	
};

#endif