/* Copyright  2014  Technicolor R&D France SNC                                            
All Rights Reserved                                                 

This program is protected by intellectual property under applicable authorship & copyright laws and international conventions. 
This program may also be subject to some patent and pending patent applications. 
You are only permitted, under Technicolor R&D France SNC owned copyrights, to reproduce, make copy, modification thereof, and reproduction of the same for research and academic purposes only. 
No other use is permitted unless expressly otherwise previously permitted in a written agreement with Technicolor R&D France SNC

Except as expressly provided herein, no license or right, express or implied, is hereby conveyed or granted by Technicolor R&D France SNC to you for any intellectual property of Technicolor R&D France SNC or its affiliated company.

YOU EXPRESSLY ACKNOWLEDGE AND AGREE THAT USE OF THE PROGRAM  IS AT YOUR SOLE RISK AND THAT THE ENTIRE RISK AS TO SATISFACTORY QUALITY, PERFORMANCE, ACCURACY AND EFFORT IS WITH YOU. TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THE PROGRAM IS PROVIDED "AS IS" AND AS AVAILABLE, WITH ALL FAULTS AND WITHOUT WARRANTY OF ANY KIND, AND TECHNICOLOR R&D FRANCE SNC HEREBY DISCLAIMS ALL WARRANTIES AND CONDITIONS WITH RESPECT TO THE PROGRAM, EITHER EXPRESS, IMPLIED OR STATUTORY, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES AND/OR  CONDITIONS OF MERCHANTABILITY, OF SATISFACTORY QUALITY, OF FITNESS FOR A PARTICULAR PURPOSE, OF ACCURACY, OF QUIET ENJOYMENT, AND NON-INFRINGEMENT OF THIRD PARTY RIGHTS. TECHNICOLOR R&D FRANCE SNC DOES NOT WARRANT AGAINST INTERFERENCE WITH YOUR ENJOYMENT OF THE PROGRAM, THAT THE FUNCTIONS CONTAINED PERFORMED BY THE PROGRAM WILL MEET YOUR REQUIREMENTS, THAT THE OPERATION OF THE PROGRAM WILL BE UNINTERRUPTED OR ERROR-FREE, OR THAT DEFECTS IN THE PROGRAM WILL BE CORRECTED NOR IMPROVED. IF ANY CORRECTION OR IMPROVEMENT OF THE PROGRAM IS MADE BY TECHNICOLOR R&D FRANCE SNC, NOTHING IN THIS LICENSE IMPLIES THAT TECHNICOLOR R&D FRANCE SNC SHALL GIVE YOU ACCESS TO SUCH CORRECTION OR IMPROVEMENT. NO ORAL OR WRITTEN INFORMATION OR ADVICE GIVEN BY TECHNICOLOR R&D FRANCE SNC OR ITS AUTHORIZED REPRESENTATIVES SHALL CREATE A WARRANTY. SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE ENTIRE COST OF ALL NECESSARY SERVICING, REPAIR OR CORRECTION. SOME JURISDICTIONS DO NOT ALLOW THE EXCLUSION OF IMPLIED WARRANTIES OR LIMITATIONS ON APPLICABLE STATUTORY RIGHTS OF A CONSUMER, SO THE ABOVE EXCLUSION AND LIMITATIONS MAY NOT APPLY TO YOU.

The use of the program implies the full acceptance of the present License.

\author Guillaume Lavou&eacute;, Xavier Rolland-Nevi&egrave;re
*/

#ifndef MATLABPOLYHEDRON_H
#define MATLABPOLYHEDRON_H

#include <CGAL/Polyhedron_incremental_builder_3.h>

/*! \file MatlabPolyhedron.h
\brief Main conversion functions for Polyhedron between MATLAB and CGAL.
\author Xavier Rolland-Nevi&egrave;re

The MatlabPolyhedron header uses a Polyhedron class, that needs to be typedef prior to 
including the header. For instance include this header after including another header containing this simple example:
\code
#include <CGAL/Simple_cartesian.h>
#include <CGAL/Polyhedron_3.h>
typedef CGAL::Simple_cartesian<double> K; 
typedef CGAL::Polyhedron_3<K> Polyhedron;
\endcode
*/

/*! \brief Procedures to convert between MATLAB and CGAL mesh representations.

MATLAB representations for meshes are based on pairs of arrays \f$(\mathbb{V},\mathbb{F})\f$, where \f$\mathbb{V}\f$ is a 3-row array
indicating the vertex locations, and \f$\mathbb{F}\f$ is a 3-row array indicating the links between vertices that form facets. 
This representation is straightforwardly taken from the Object File Format, whereas CGAL uses a halfedge data structure http://doc.cgal.org/latest/HalfedgeDS/index.html.

This namespace provides 2 conversion routines:
-# conversion from MATLAB arrays to CGAL polyhedron: Utils::MATLAB2Polyhedron;
-# conversion from CGAL polyhedron to already allocated MATLAB arrays: Utils::Polyhedron2MATLAB.

Conversions from MATLAB to CGAL rely on a specialized polyhedron builder class
http://doc.cgal.org/latest/Polyhedron/classCGAL_1_1Polyhedron__incremental__builder__3.html. 
The Polyhedron class is a prevously defined type inheriting from CGAL::Polyhedron_3<K> (see MatlabPolyhedron.h), a.k.a. a CGAL polyhedron.
*/
namespace Utils
{
	/*! \brief Conversion from MATLAB mesh representation to CGAL polyhedron representation.
	\param *mxV Pointer to the array of vertex locations.
	\param *mxF Pointer to the array of facets.
	\return A pointer to a CGAL polyhedron.

	MATLAB mesh representation consists of two matrices listing the vertex locations and
	the facets. The return type must be a preset typedef. To avoid deallocation issues, the instance
	of the return object is wrapped in a unique_ptr (see http://www.cplusplus.com/reference/memory/unique_ptr/).

	<i>NB</i>: MATLAB indices start at 1. Conversions automatically take care of the offset with C++.*/
	std::unique_ptr<Polyhedron> MATLAB2Polyhedron(const mxArray *mxV, const mxArray *mxF);

	/*!	\brief Conversion from CGAL polyhedron to a MATLAB-compatible representation of the mesh.
	\param polyhedron CGAL polyhedron to be converted.
	\param points Pointer to an array *already allocated* where the vertex locations are stored.
	\param facets Pointer to an array *already allocated* where the facets are stored.

	Because MATLAB systematically uses double type, facet indices and vertex locations are expressed
	with double. Calls to this function are responsible for the allocation and destruction of both
	arrays filled. Use polyhedron member functions size_of_vertices() and size_of_facets(). For instance, using
	a pointer pP1 to a polyhedron object:

	\code
	Polyhedron* pP1;
	mxArray* Mpoints = mxCreateDoubleMatrix(3, pP1->size_of_vertices(), mxREAL);
	mxArray* MFacets = mxCreateDoubleMatrix(3, pP1->size_of_facets(), mxREAL);
	Utils::Polyhedron2MATLAB(*pP1, mxGetPr(Mpoints), mxGetPr(MFacets));
	\endcode

	<i>NB</i>: MATLAB indices start at 1. Conversions automatically take care of the offset with C++.*/
	void Polyhedron2MATLAB(const Polyhedron& polyhedron, double* points, double* facets); 

	/*! \brief Builder class for a triangle surface mesh based on a list of facets and vertices.*/
	template<class HDS>
	class polyhedron_builder : public CGAL::Modifier_base<HDS> 
	{
	private:
		std::vector<double> &coords;
		std::vector<int>    &tris;
	public:
		/*! \brief Construct a polyhedron builder class from a series of vertex coordinates and a series of triangle facets
			\param _coords	Vector of concatenated vertex coordinates
			\param _tris Vector of concatenated facet indices

			The coordinates are concatenated as /f$x_0,y_0,z_0,x_1,...$/f. The facet indices are concatenation of integer
			triplets linking 3 vertices together to form a facet.
		*/
		polyhedron_builder(std::vector<double> &_coords, std::vector<int> &_tris ) : coords(_coords), tris(_tris) {}
		/*! \brief Perform the construction of the triangle surface mesh.*/
		void operator()( HDS& hds) 
		{
			// create a cgal incremental builder
			CGAL::Polyhedron_incremental_builder_3<HDS> B(hds, true);
			B.begin_surface(coords.size() / 3, tris.size() / 3); 
			// add the polyhedron vertices
			for (int i = 0; i < (int) coords.size(); i+=3)
			{
				B.add_vertex(HDS::Vertex::Point(coords[i+0], coords[i+1], coords[i+2]));
			}
			// add the polyhedron triangles
			for(int i = 0; i < (int) tris.size(); i+=3)
			{
				B.begin_facet();
				B.add_vertex_to_facet(tris[i+0]);
				B.add_vertex_to_facet(tris[i+1]);
				B.add_vertex_to_facet(tris[i+2]);
				B.end_facet();
			} 
			// finish up the surface
			B.end_surface();
		}
	};

	void Polyhedron2MATLAB(const Polyhedron& polyhedron, double* points, double* facets)
	{
		Polyhedron::Point_const_iterator pit = polyhedron.points_begin();
		CGAL_For_all(pit, polyhedron.points_end())
		{
			*points++ = pit->x();
			*points++ = pit->y();
			*points++ = pit->z();
		}
		CGAL::Inverse_index<Polyhedron::Vertex_const_iterator> index(polyhedron.vertices_begin(), polyhedron.vertices_end());
		Polyhedron::Facet_const_iterator fit = polyhedron.facets_begin();
		CGAL_For_all(fit, polyhedron.facets_end())
		{
			auto hc = fit->facet_begin();
			auto hc_end = hc;
			std::size_t n = circulator_size(hc);
			CGAL_assertion(n == 3);
			do
			{
				*facets++ = static_cast<double>(1 + index[Polyhedron::Vertex_const_iterator(hc->vertex())]);
				++hc;
			}
			while(hc != hc_end);
		}
	}

	std::unique_ptr<Polyhedron> MATLAB2Polyhedron(const mxArray *mxV, const mxArray *mxF)
	{
		double* V = (double*) mxGetData(mxV);	
		size_t nbv = mxGetN(mxV);
		std::vector<double> coords(V, V + 3 * nbv);
		size_t nbf = mxGetN(mxF);
		double* F = (double*) mxGetData(mxF);
		std::vector<int> ids;
		for (size_t i = 0 ; i < nbf ; i++)
		{
			ids.push_back(static_cast<int>(*F++ - 1));
			ids.push_back(static_cast<int>(*F++ - 1));
			ids.push_back(static_cast<int>(*F++ - 1));
		}
		std::unique_ptr<Polyhedron> pP = std::unique_ptr<Polyhedron>(new Polyhedron);
		polyhedron_builder<Polyhedron::HalfedgeDS> builder(coords, ids);
		pP->delegate(builder);
		return pP;
	}

}
#endif